#**********************************************************************************************************+++++++++++++++*********************!
# (c) nextnano GmbH                                                                                                                            !
#                                                                                                                                              !
# This is ascript for the following tutorial:                                                                                                  !
#   at https://www.nextnano.com/documentation/tools/nextnanomat/export_functionalities/paraview/python_2d.html#nnmt-export-paraview-python-2d  !
#                                                                                                                                              !
# For comments or feedback, please contact support@nextnano.com                                                                                !
#**********************************************************************************************************************************************!

from paraview.simple import *
from os import makedirs, getcwd
from os.path import isdir, basename, join, normpath, exists
from paraview import servermanager
  
 
#CONFIGURATION PARAMETERS ( Do not change this section: the settings will be modified through the GUI)
WARP = False
OVERLAY = False
                          
SOLID_COLOR = False 
CAMERA_VIEW_UP_Y = False 
MAXIMUM_NUMBER_DATASETS = False
OPACITY = False

OPACITY_VALUE = 0.22
MAXIMUM_NUMBER_DATASETS_VALUE = 8

fileList = IMPORTED_FILE_LIST
OVERLAY_FILE = IMPORTED_OVERLAY_FILE
IMAGE_PREFIX = 'Image'  
                                 
IMAGE_DIRECTORY = IMPORTED_IMAGE_DIRECTORY   
# End of the CONFIGURATION section


#Default values of this script (not changed externally)
OVERLAY_PALETTE = [0.5764705882352941, 0.7607843137254902, 0.6470588235294118, 0.9882352941176471, 0.5529411764705883, 0.3843137254901961, 0.5529411764705883, 0.6274509803921569, 0.796078431372549]
PLOT_PALETTE = 'Rainbow Uniform'
OPACITY_VALUE_DEFAULT = 0.22
MAXIMUM_NUMBER_DATASETS_DEFAULT = 8

DEFAULT_IMAGE_DIR = 'C:/ParaView Images/'

CAMERA_POSITION = [-42, -4.76, 15.7]
CAMERA_VIEW_UP = [0.18, 0.6285, 0.7536]

SHOW_COLOR_BAR = False
RESCALE_OPTION = False    

   
def createRenderView():
    """Create and initialize renderView (all files in only one window)"""
    
    renderView = GetActiveViewOrCreate('RenderView')
    renderView.Background = [1,1,1]
    
    return renderView


def createDisplay(renderView, object, curve, rescale, colorTheme= False):
    """Create and initialize display. Display behaves like a controller of each specific object"""
    
    display = Show(object, renderView)
    display.Representation = 'Surface'    
    display.AmbientColor = [0.0, 0.0, 0.0]
    
    if colorTheme:
        ColorBy(display, None)
        display.AmbientColor = [0.784, 1.0, 0.541]
        display.DiffuseColor = [0.784, 1.0, 0.541]
    else:    
        ColorBy(display, ('POINTS', curve), separate = True)
        curveLUT = GetColorTransferFunction(curve, display)
        curveLUTColorBar = GetScalarBar(curveLUT, renderView)
        curveLUTColorBar.AutomaticLabelFormat = 0
        curveLUTColorBar.LabelFormat = '%-#5.3f'
        curveLUTColorBar.RangeLabelFormat = '%-#5.3f'
        curveLUTColorBar.ScalarBarLength = 0.35
        curveLUTColorBar.Position = [0.72, 0.24]
        curveLUTColorBar.Title = curve.split('[')[0] 
        curveLUTColorBar.ComponentTitle = ''       
        curveLUTColorBar.TitleColor = [0.0, 0.0, 0.0]
        curveLUTColorBar.LabelColor = [0.0, 0.0, 0.0]        
        curveLUT.ApplyPreset(PLOT_PALETTE, True)        
        display.SetScalarBarVisibility(renderView, SHOW_COLOR_BAR)
      
    return display


def getImageDir(imageDirectory):
    """Set image directory to the default, if the path is empty, and create if it does not exist"""
    
    if imageDirectory == '':
        imagePath = DEFAULT_IMAGE_DIR
    else:
        imagePath = imageDirectory    
        
    if not exists(imagePath):
        makedirs(imagePath)
        print('Created directory for ParaView images at ' + imagePath)
        
    return imagePath


def getCurveList(plotFile):
    """Get a list with all labels of the datasets of plotFile (all data that it is not coordinate).
       The number of elements of the list will be reduced according the parameter MAXIMUM_NUMBER_DATASETS_VALUE 
       if it is less the lenght of the list.
       Additionally, in files whose filename contains the string shift, the eigenvalues will be excluded of the 
       the list of curves to be plotted"""
    
    curveList = plotFile.PointArrayStatus
    curveName = plotFile.FileName

    if "_shift" in curveName[0]:
        curveList = [ label for label in plotFile.PointArrayStatus if (not "E_" in label) and (not "[eV]" in label)]
    else:
        curveList = plotFile.PointArrayStatus
    
    if len(curveList) > MAXIMUM_NUMBER_DATASETS_VALUE:
        curveList = curveList[0:MAXIMUM_NUMBER_DATASETS_VALUE]

    print("Curve list to be actually plotted: ")
    print(curveList)    
    
    return curveList    


def getWarpScalar(plotFile, curve):
    """Compute the value of warpScalar according the range of the dataset. Usually is called using the first curve 
    of plot file"""
    
    dataRange = plotFile.PointData.GetArray(curve).GetRange() 
    bounds = plotFile.GetDataInformation().GetBounds()
    
    averageBounds = bounds[1] + bounds[3] - bounds[0] + bounds[2]
    
    maximumRange = dataRange[1] - dataRange[0]
    
    if maximumRange != 0:
        _warpScalar = 0.4 * averageBounds/ maximumRange
        print("Warp scalar: " + str(_warpScalar))
        translateInsteadWarp = False
    else:
        _warpScalar = 0
        translateInsteadWarp = True
        print("No warped images generated, because this is a plane")    
        
    if 'bandedge' in plotFile.FileName[0]:
        _warpScalar = 1.0
        
    return _warpScalar
 
 
def loadOriginalFile(plotObject, isOverlay=False):
    """ Load the datafile and insert it in the Pipeline Browser"""
    
    if isOverlay:
        return XMLRectilinearGridReader(registrationName= "OVERLAY: " + plotObject, FileName=plotObject)
    else:
        return XMLRectilinearGridReader(registrationName= plotObject, FileName=plotObject)
        
 
def plotComponent(plotFile, curve, rescale):
    """Extract component from the plotFile"""
    
    print('\n\nProcessing curve ' + curve)
    curveName = curve.replace('[', ' (').replace(']', ')')
    
    componentFile = ExtractComponent(registrationName= curveName ,Input=plotFile)  
    componentFile.InputArray = ['POINTS', curve]
    componentFileDisplay = createDisplay(renderView, componentFile, curve, rescale, colorTheme = False)
    renderView.CameraViewUp = [0, 1, 0]
    renderView.InteractionMode = '2D'
    renderView.Update()
    
    return componentFile


def plotWarpedFile(curve, componentFile, warpScalarValue, warpScalarOverlay=0.0, shifted= False, isOverlay=False, overlayColor= 0, plotOpaque=False): 
    """Warp a 2D map by warpScalarValue. 
    If shifted is True, translates the plot according the value of warpScalar of the overlay and the eigenvalue.
    If the curve to be plotted is larger than the number of colors in the palette, it will be plotted with the last color 
    in the palette"""
    
    overlayPalette = OVERLAY_PALETTE
    translateInsteadWarp = False
    rescale = RESCALE_OPTION
    translationValue = 0.0
    
    warpedFileName = curve.split('[')[0] + '_warped'
    if isOverlay:
        warpedFileName = 'overlay_' + warpedFileName                        
        rescale = False
    
    renderView.Update()
    
    #Valid only for shifted probabilities
    if shifted : 
        correspondentEnergy = curve.replace('Psi^2_', 'E_').replace('nm^-2','eV')
        eigenvalue = plotFile.PointData[correspondentEnergy].GetRange()[0]
        
        calculatorFileName = warpedFileName.replace('_warped','_no_shifted')
        calculatorFile = Calculator(registrationName= calculatorFileName , Input=componentFile)
        calculatorFile.Function = '\"' + curve + '\"-\"' + correspondentEnergy + '\"'
        calculatorFile.ResultArrayName = 'Only_' + curve.split('[')[0]
        
        calculatorResult = calculatorFile.ResultArrayName
        calculatorFileDisplay = createDisplay(renderView, calculatorFile, curve, rescale, colorTheme = SOLID_COLOR)
        
        warpedFile = WarpByScalar(registrationName=warpedFileName, Input=calculatorFile)
        curve = calculatorResult

        translationValue = eigenvalue
        Hide(calculatorFile)
    else:
        warpedFile = WarpByScalar(registrationName= warpedFileName, Input=componentFile)
    
    warpedFile.Scalars = ['POINTS', curve]
    warpedFile.ScaleFactor = warpScalarValue        

    # setting display
    colorTheme = SOLID_COLOR
    if plotOpaque:
        colorTheme = True
        
    warpedFileDisplay = createDisplay(renderView, warpedFile, curve, rescale, colorTheme = colorTheme)
 
    if OPACITY or plotOpaque:
        warpedFileDisplay.Opacity = OPACITY_VALUE
    warpedFileDisplay.Position = [0.0, 0.0, translationValue]
    warpedFileDisplay.Scale = [1.0,1.0,1.0] 
    
    if isOverlay:           # Special Palette for overlay
        ColorBy(warpedFileDisplay, None)    
        warpedFileDisplay.Opacity = OPACITY_VALUE_DEFAULT   
        
        numberColorsInPalette = len(overlayPalette)
        if overlayColor > numberColorsInPalette:
            overlayColor = numberColorsInPalette - 1
            
        warpedFileDisplay.AmbientColor = [overlayPalette[overlayColor], 0.9803921568627451/(overlayColor + 1), 1.0]
        warpedFileDisplay.DiffuseColor = [overlayPalette[overlayColor], 0.9803921568627451/(overlayColor + 1), 1.0]
        
        if "Fermi_level" in curve:
            ColorBy(warpedFileDisplay, None) 
            warpedFileDisplay.Opacity = 0.9
            warpedFileDisplay.AmbientColor = [1.0, 0.9059, 0.8314]
            warpedFileDisplay.DiffuseColor = [1.0, 0.9059, 0.8314]
            if 'hole' in curve:
                warpedFileDisplay.AmbientColor = [1.0, 0.9059, 0.8314]
                warpedFileDisplay.DiffuseColor = [1.0, 0.9059, 0.8314]                 
           
    Show(componentFile)
    Hide(componentFile)
    
    return warpedFile

        
def addPropertyLink(objectList, property):
    """Link property of all components of the data file"""

    propertyLink = servermanager. vtkSMPropertyLink()
    link = servermanager.vtkSMProxyLink()
    
    for  item in objectList:
         firstObjectProxy = item[0]
         secondObjectProxy = item[1]
         firstObject = proxyManager.GetProxyName("sources", firstObjectProxy)
         secondObject = proxyManager.GetProxyName("sources", secondObjectProxy)

         link.AddLinkedProxy(firstObjectProxy.SMProxy,  1)
         link.AddLinkedProxy(secondObjectProxy.SMProxy,  2)    

         propertyLink.AddLinkedProperty(firstObjectProxy.SMProxy, property, 1)
         propertyLink.AddLinkedProperty(secondObjectProxy.SMProxy, property, 2)    
         propertyLink.AddLinkedProperty(firstObjectProxy.SMProxy, property, 2)
         propertyLink.AddLinkedProperty(secondObjectProxy.SMProxy, property, 1)
         
         suffix = '_' + firstObject.replace('_warped','') + '_' + secondObject.replace('_warped','')
         
         proxyManager.RegisterLink(property + suffix, propertyLink)
         proxyManager.RegisterLink("link" + suffix , link)


def isShifted(file):
    """Verifies if file is shifted"""    

    if "_shift" in basename(file):
        return True
    else:
        return False
        

#---------------------- INITIALIZATION --------------------------------------------------------------------#

ResetSession()
paraview.simple._DisableFirstRenderCameraReset()

renderView = createRenderView()    

proxyManager = servermanager.ProxyManager()

imageDir = getImageDir(IMAGE_DIRECTORY)
imagePrefix = IMAGE_PREFIX


#---------------------- PLOT OF THE OVERLAY FILE ----------------------------------------------------------#

imageIndex = 0
warpScalarOverlay = 1

firstPlot = True
if OVERLAY:
    print("\n++++++++++++++++++++\nProcessing overlay file : " + OVERLAY_FILE)
    overlayShifted = isShifted(OVERLAY_FILE) 
    rescale = True

    plotFile = loadOriginalFile(OVERLAY_FILE)
    Hide(plotFile)
    curveList = getCurveList(plotFile)
    
    print("Overlay plot options: ")
    print(curveList)
    
    if overlayShifted and "bandedge" in basename(OVERLAY_FILE):
        warpScalar = 1.0
        plotOpaque = True
    else:
        warpScalar = getWarpScalar(plotFile, curveList[0])
        plotOpaque = False

    for curveIndex in range(0,len(curveList)):
        curve = curveList[curveIndex]
        
        firstCurve = curveIndex == 0
        
        if firstCurve:
            rescaleCurve = True    
        else:
            rescaleCurve = RESCALE_OPTION                      
        componentFile = plotComponent(plotFile, curve, rescale)
        
        if WARP:
            if firstCurve:
                warpScalarOverlay = getWarpScalar(plotFile, curveList[0])
                print("Warp scalar from overlay: " + str(warpScalarOverlay))            
        
            warpedFile = plotWarpedFile(curve, componentFile, warpScalarOverlay, shifted= overlayShifted, isOverlay=True, overlayColor = curveIndex, plotOpaque= True)
            
            renderView.ResetCamera()
            if firstCurve:
                firstOverlayCurve = curve.split('[')[0]  + "_warped"
            else:
                connection = [previousObject, warpedFile]
                addPropertyLink([connection],"ScaleFactor")

            previousObject = warpedFile

            if not overlayShifted:
                Hide(warpedFile)

            Show(warpedFile)
                       
            renderView.OrientationAxesVisibility = 0
            renderView.Update()            
else:
    overlayShifted = False

#---------------------- PLOT OF ALL FILES  ----------------------------------------------------------#

firstPlot = True

for plotObject in fileList:  
    print("\n++++++++++++++++++++\nProcessing file : " + plotObject)
    shiftedFile = "_shift" in basename(plotObject)
    
    #Rescale at least the first plotFile
    if firstPlot:
        rescale = True     

    plotFile = loadOriginalFile(plotObject)
    Hide(plotFile)
 
    curveList = getCurveList(plotFile)
 
    if overlayShifted and "bandedge" in basename(plotObject):
        plotOpaque = True
    else:
        plotOpaque = False
        
    warpScalar = getWarpScalar(plotFile, curveList[0])
    
    for curveIndex in range(0,len(curveList)):             
        curve = curveList[curveIndex]
        firstCurve = curveIndex == 0
        
        if firstCurve:
            rescaleCurve = True    
        else:
            rescaleCurve = RESCALE_OPTION
           
        componentFile = plotComponent(plotFile, curve, rescale) 
  
        if WARP:      
            warpedFile = plotWarpedFile(curve, componentFile, warpScalar, warpScalarOverlay=warpScalarOverlay, shifted=shiftedFile, plotOpaque= plotOpaque)

            if not firstCurve:
                connection = [previousObject, warpedFile]
                addPropertyLink([connection],"ScaleFactor")
            else:
                renderView.ResetCamera()

            previousObject = warpedFile
            
            Hide(plotFile)
            Hide(componentFile)
            Show(warpedFile)

        renderView.Update()
        renderView.OrientationAxesVisibility = 0
        
        rescale = RESCALE_OPTION
        
    firstPlot = False   

print("\n\n++++++++++++++++++++END OF THE POST-PROCESSING SCRIPT")

if WARP:
    renderView.InteractionMode = '3D'
    if CAMERA_VIEW_UP_Y:
        renderView.ResetActiveCameraToPositiveY()
    else:
        renderView.CameraViewUp = CAMERA_VIEW_UP
        renderView.CameraPosition = CAMERA_POSITION  
else:
    renderView.InteractionMode = '2D'
    renderView.ResetActiveCameraToPositiveZ()

ResetCamera()
